/*
 * effects.c
 *
 *  Created on: 21-01-2012
 *      Author: Andrzej Michalec (amichalec.net)
 */

#include <stdlib.h>
#include <util/delay.h>
#include "snowflake.h"
#include "effects.h"

light_effect_t * eff;
light_effect_t * eff2;

void __wait(uint8_t seconds) {
	for (uint8_t i = 0; seconds == 0 || i < seconds; i++) {
		_delay_ms(1000);
	}
}

light_effect_t * __shine_effect() {
	const uint8_t SIZE = 40;
	light_effect_t * pe = sf_allocate(SIZE);
	for (uint8_t i = 0; i < 16; i++) {
		*(pe->brightness_tab + i) = i;
		*(pe->time_tab + i) = 10 * (16 - i);
	}
	for (uint8_t i = 16; i < 24; i++) {
		int m = i % 3;
		if (m == 0) {
			*(pe->brightness_tab + i) = 12;
			*(pe->time_tab + i) = 100;
		} else if (m == 1) {
			*(pe->brightness_tab + i) = 15;
			*(pe->time_tab + i) = 10;
		} else {
			*(pe->brightness_tab + i) = 14;
			*(pe->time_tab + i) = 30;
		}
	}
	for (uint8_t i = 24; i < 40; i++) {
		*(pe->brightness_tab + i) = 39 - i;
		*(pe->time_tab + i) = 10 * (i - 23);
	}
	return pe;
}

void run_random_light() {
	uint8_t i = rand() % SF_MAX_LIGHTS;
	while (sf_is_effect_in_progress(i)) {
		i++;
		i %= SF_MAX_LIGHTS;
	}
	sf_set_light_effect_data(SF_LIGHT_TYPE_POINT, i, eff);
}

void __listener(uint8_t light) {
	run_random_light();
}

void run_shine_effect(uint8_t seconds, uint8_t lightsAtOnce) {
	eff = __shine_effect();
	sf_set_listener(__listener);
	if (lightsAtOnce > SF_MAX_LIGHTS) {
		lightsAtOnce = SF_MAX_LIGHTS;
	}
	for (uint8_t i = 0; i < lightsAtOnce; i++) {
		run_random_light();
		for (int j = 0; j < 1 + rand() % 5; j++) {
			_delay_ms(150);
		}
	}
	__wait(seconds);
	sf_set_listener(NULL); // fade out
	_delay_ms(4000);
	sf_free(eff);
}

//---------------------------------------------------------

light_effect_t * __pulse_effect() {
	const uint8_t SIZE = 4;
	light_effect_t * pe = sf_allocate(SIZE);
	*(pe->brightness_tab + 0) = 3;
	*(pe->time_tab + 0) = 3000;
	*(pe->brightness_tab + 1) = 0;
	*(pe->time_tab + 1) = 100;
	*(pe->brightness_tab + 2) = 6;
	*(pe->time_tab + 2) = 100;
	*(pe->brightness_tab + 3) = 3;
	*(pe->time_tab + 3) = 3000;
	pe->next_effect = pe;
	return pe;
}

void run_pulse_star_effect(uint8_t seconds) {
	eff = __shine_effect();
	eff->next_effect = eff;
	eff2 = sf_white2rgb(eff);
	eff2->next_effect = eff2;
	sf_set_light_effect_data(SF_LIGHT_TYPE_POINT, SF_RGB_LIGHT_NUMBER, eff2);
	_delay_ms(150);
	for (uint8_t i = 0; i < SF_MAX_CIRCLES; i++) {
		sf_set_light_effect_data(SF_LIGHT_TYPE_CIRCLE, i, eff);
		_delay_ms(150);
	}
	__wait(seconds);
	eff2->next_effect = NULL;
	eff->next_effect = NULL;
	_delay_ms(2500);
	for (uint8_t i = 0; i < SF_MAX_CIRCLES; i++) {
		sf_set_light(SF_LIGHT_TYPE_CIRCLE, i, 0);
	}
	sf_free(eff2);
	sf_free(eff);
}

//---------------------------------------------------------

/**
 * Triangular wave with values 0..15 raising up and falling down within
 * give periodLenght. Step is incremental counter that can span
 * any number of periods.
 */
uint8_t __triangular_wave(uint16_t step, uint8_t periodLength) {
	uint8_t instep = step % periodLength;
	uint8_t up = (instep < (periodLength / 2));
	if (up) {
		return 2 * (SF_MAX_BRIGHTNESS + 1) * instep / periodLength;
	} else {
		return 2 * (SF_MAX_BRIGHTNESS + 1) * (periodLength - instep - 1)
				/ periodLength;
	}
}

light_effect_t * __rainbow_effect() {
	const uint8_t SIZE = 60;
	const uint8_t PR = SIZE / 5;
	const uint8_t PG = SIZE / 3;
	const uint8_t PB = SIZE / 4;
	light_effect_t * pe = sf_allocate(SIZE);
	for (uint16_t i = 0; i < SIZE; i++) {
		uint8_t r = __triangular_wave(i, PR);
		uint8_t g = __triangular_wave(i, PG);
		uint8_t b = __triangular_wave(i, PB);
		*(pe->brightness_tab + i) = SF_RGB(r,g,b);
		*(pe->time_tab + i) = 50;
	}
	return pe;
}

void run_rainbow_effect(uint8_t seconds) {
	eff = __rainbow_effect();
	eff->next_effect = eff;
	for (uint8_t j = 0; j < 10; j++) {
		for (uint8_t i = 0; i < SF_MAX_CIRCLES; i++) {
			sf_set_light(SF_LIGHT_TYPE_CIRCLE, i, 4 * i * j / 10);
		}
		_delay_ms(100);
	}
	sf_set_light_effect_data(SF_LIGHT_TYPE_POINT, SF_RGB_LIGHT_NUMBER, eff);
	__wait(seconds);
	eff->next_effect = NULL;
	_delay_ms(1500);
	sf_set_light(SF_LIGHT_TYPE_POINT, SF_RGB_LIGHT_NUMBER, 0);
	for (uint8_t j = 0; j < 10; j++) {
		for (uint8_t i = 0; i < SF_MAX_CIRCLES; i++) {
			sf_set_light(SF_LIGHT_TYPE_CIRCLE, i, 4 * i * (9 - j) / 10);
		}
		_delay_ms(100);
	}
	sf_free(eff);
}

//---------------------------------------------------------

light_effect_t * __fadeout_effect(uint16_t in_delay, uint16_t out_delay,
		uint16_t dark_delay) {
	const uint8_t SIZE = 23;
	light_effect_t * pe = sf_allocate(SIZE);
	uint8_t i = 0;
	for (; i < 7; i++) {
		*(pe->brightness_tab + i) = 1 + 2 * i;
		*(pe->time_tab + i) = in_delay;
	}
	for (; i < SIZE; i++) {
		*(pe->brightness_tab + i) = SIZE - 1 - i;
		*(pe->time_tab + i) = out_delay;
	}
	*(pe->time_tab + SIZE - 1) = dark_delay;
	return pe;
}

void run_windmill_effect(uint8_t seconds) {
	eff = __fadeout_effect(20, 100, 700);
	eff->next_effect = eff;
	sf_set_light(SF_LIGHT_TYPE_POINT, SF_RGB_LIGHT_NUMBER, SF_RGB(0, 2, 2));
	for (uint8_t i = 0; i < SF_MAX_ARMS; i++) {
		sf_set_light_effect_data(SF_LIGHT_TYPE_ARM, i, eff);
		_delay_ms(400);
	}
	__wait(seconds);
	eff->next_effect = NULL;
	sf_set_light(SF_LIGHT_TYPE_POINT, SF_RGB_LIGHT_NUMBER, 0);
	_delay_ms(2000);
	sf_free(eff);
}

//---------------------------------------------------------

void run_atom_effect(uint8_t seconds) {
	eff = __fadeout_effect(20, 30, 500);
	eff->next_effect = NULL;
	uint8_t lines[3][8] = { { 0, 1, 2, 3, 15, 14, 13, 12 }, { 4, 5, 6, 7, 19,
			18, 17, 16 }, { 8, 9, 10, 11, 23, 22, 21, 20 } };
	int idx[3] = { 0, 5, 3 };
	int dir[3] = { 1, 1, -1 };
	const int DELAY = 100;
	int count = seconds * 1000 / DELAY;
	for (uint8_t i = 0; i < SF_MAX_BRIGHTNESS; i++) {
		sf_set_light(SF_LIGHT_TYPE_POINT, SF_RGB_LIGHT_NUMBER, SF_RGB(i, i, 0));
		_delay_ms(20);
	}
	for (; count > 0; count--) {
		for (uint8_t i = 0; i < 3; i++) {
			idx[i] += dir[i];
			if (idx[i] > 7) {
				idx[i] = 6;
				dir[i] = -1;
			}
			if (idx[i] < 1) {
				idx[i] = 0;
				dir[i] = 1;
			}
			sf_set_light_effect_data(SF_LIGHT_TYPE_POINT, lines[i][idx[i]], eff);
		}
		_delay_ms(DELAY);
	}
	for (uint8_t i = SF_MAX_BRIGHTNESS - 1; i > 0; i--) {
		sf_set_light(SF_LIGHT_TYPE_POINT, SF_RGB_LIGHT_NUMBER, SF_RGB(i, i, 0));
		_delay_ms(20);
	}
	sf_set_light(SF_LIGHT_TYPE_POINT, SF_RGB_LIGHT_NUMBER, 0);
	_delay_ms(2000);
	sf_free(eff);
}
